import torch
import torch.nn as nn
import torch.nn.functional as F
from abc import abstractclassmethod

from torch.nn.modules.conv import Conv1d

class Lambda(nn.Module):
    """Wrap a function as `nn.Module`."""
    def __init__(self, fn):
        super(Lambda, self).__init__()
        self.lambda_func = fn

    def forward(self, *args):
        return self.lambda_func(*args)


class SamePadding(nn.Module):
    """
    Pad the output of a module to have the same shape as its input.

    Args:

    * `module` (nn.Module): Target module, usually a convlutional one.

    * `n_dims` (int): Number of dimensions to be pad, if `None` will be infered automatically.

    Usage:
    >>> conv2d = SamePadding(
    ...     nn.Conv2d(16, 32, 3)
    ... )
    >>> input = torch.randn(1, 16, 7, 5)
    >>> output = conv2d(input)
    >>> # The output shape will be equal to the input shape along last `n_dims`
    >>> print(input.shape, output.shape)

    Note:

    If the convlution module has a `stride` > 1, the result may not be what you expect.
    """

    dims_dict = {nn.Conv1d:1, nn.Conv2d:2, nn.Conv3d:3, nn.Linear:1}

    def __init__(self, module, n_dims=None):
        super(SamePadding, self).__init__()
        self.module = module
        self.n_dims = n_dims if n_dims!=None else self.dims_dict[type(module)]
        self.pad_list = None

    def build_pad_list(self, diff_shape):
        self.pad_list = []
        for i in range(self.n_dims):
            d = diff_shape[i]
            if d % 2 == 0:
                self.pad_list += [d//2, d//2]
            else:
                self.pad_list += [d//2, d//2+1]
        self.pad_list = tuple(self.pad_list)

    def forward(self, x):

        def fwd_build(x):
            i_shape = torch.tensor(x.shape)
            x = self.module(x)
            o_shape = torch.tensor(x.shape)
            diff_shape = list(reversed(i_shape - o_shape))
            self.build_pad_list(diff_shape)
            x = F.pad(x, self.pad_list)
            return x

        if self.pad_list == None:
            return fwd_build(x)

        x = self.module(x)
        x = F.pad(x, self.pad_list)
        return x

class KerasLayer(nn.Module):
    def __init__(self, *args, **kwargs) -> None:
        super(KerasLayer, self).__init__()
        self.args = args
        self.kwargs = kwargs
        self.module = None

    @abstractclassmethod
    def build(self, in_shape: torch.Size):
        pass

    def forward(self, x):
        if not self.module:
            self.module = self.build(x.shape)
        return self.module.forward(x)

# Conv
class Conv1d(KerasLayer):
    def build(self, in_shape):
        return nn.Conv1d(in_shape[1], *self.args, **self.kwargs)

class Conv2d(KerasLayer):
    def build(self, in_shape):
        return nn.Conv2d(in_shape[1], *self.args, **self.kwargs)

class Conv3d(KerasLayer):
    def build(self, in_shape):
        return nn.Conv3d(in_shape[1], *self.args, **self.kwargs)

# BatchNorm
class BatchNorm1d(KerasLayer):
    def build(self, in_shape):
        return nn.BatchNorm1d(in_shape[1], *self.args, **self.kwargs)

class BatchNorm2d(KerasLayer):
    def build(self, in_shape):
        return nn.BatchNorm2d(in_shape[1], *self.args, **self.kwargs)

class BatchNorm3d(KerasLayer):
    def build(self, in_shape):
        return nn.BatchNorm3d(in_shape[1], *self.args, **self.kwargs)

class Linear(KerasLayer):
    def build(self, in_shape):
        return nn.Linear(in_shape[1], *self.args, **self.kwargs)


import torch
import time
import pandas as pd
from enum import Enum

class Events(Enum):
    ON_EPOCH_END = 'on_epoch_end'
    ON_EPOCH_BEGIN = 'on_epoch_begin'
    ON_TRAIN_BEGIN = 'on_train_begin'
    ON_TRAIN_END = 'on_train_end'


class StopTrainingError(Exception):
    def __init__(self):
        pass

class Trainer(object):
    def __init__(self, **kwargs) -> None:
        for key, value in kwargs.items():
            setattr(self, key, value)
        self.logger = Logger(self)
    
    def register_callbacks(self, callbacks):
        self.event_dict = {Events(k).value: list() for k in Events}

        if not isinstance(callbacks, list):
            callbacks = [callbacks]
        for c in callbacks:
            for key, func in c.get_callbacks_dict().items():
                self.event_dict[Events(key).value] += [func]
    
    def __fire_event(self, key):
        func_list = self.event_dict[Events(key).value]
        for func in func_list:
            func(self)

    def run(self, train_loader, val_loader, max_epochs, verbose, precise_train_metrics):
        self.max_epochs = max_epochs
        self.logger.verbose = verbose
        self.logger.on_train_begin(train_loader, val_loader)
        self.__fire_event(Events.ON_TRAIN_BEGIN)

        if precise_train_metrics:
            train_fn = self.train_precise_mode
        else:
            train_fn = self.train_fast_mode

        for epoch in range(1, max_epochs+1):
            self.epoch = epoch
            self.logger.on_epoch_begin()
            self.__fire_event(Events.ON_EPOCH_BEGIN)

            train_metrics = train_fn(train_loader)
            val_metrics = self.evaluate(val_loader) if val_loader else {}

            self.logger.on_epoch_end(epoch=epoch, max_epochs=max_epochs, train_metrics=train_metrics, val_metrics=val_metrics)

            try:
                self.__fire_event(Events.ON_EPOCH_END)
            except StopTrainingError:
                break

        self.__fire_event(Events.ON_TRAIN_END)
        return self.logger.history

    def train_precise_mode(self, data_loader):
        self.model.train()

        has_w = len(data_loader.dataset[0]) == 3

        for t in data_loader:
            x = t[0].to(device=self.device)
            y = t[1].to(device=self.device)
            self.optimizer.zero_grad()
            y_pred = self.model.forward(x)

            loss = self.loss(y_pred, y)

            if has_w:
                if loss.dim() == 0:
                    raise ValueError("You should set the loss function with `reduction='none'` when using sample_weight.")
                t[2] = t[2].to(device=self.device)
                loss = loss * t[2]

            loss.mean().backward()
            self.optimizer.step()
        
        return self.evaluate(data_loader)

    def train_fast_mode(self, data_loader):
        self.model.train()
        metrics, y_true, y_pred = {}, [], []
        sample_weight = []

        has_w = len(data_loader.dataset[0]) == 3

        for t in data_loader:
            x = t[0].to(device=self.device)
            y = t[1].to(device=self.device)
            self.optimizer.zero_grad()
            y_batch_pred = self.model.forward(x)

            loss = self.loss(y_batch_pred, y)

            if has_w:
                if loss.dim() == 0:
                    raise ValueError("You should set the loss function with `reduction='none'` when using sample_weight.")
                t[2] = t[2].to(device=self.device)
                loss = loss * t[2]
                sample_weight.append(t[2])

            loss.mean().backward()
            self.optimizer.step()

            y_pred.append(y_batch_pred.detach())
            y_true.append(y)
            
        y_true = torch.cat(y_true)
        y_pred = torch.cat(y_pred)

        if has_w:
            sample_weight = torch.cat(sample_weight)

        with torch.no_grad():
            for key, score_fn in self.metrics.items():
                if has_w and 'loss' in key:
                    loss = score_fn(y_pred, y_true)
                    if loss.dim() == 0:
                        raise ValueError("You should set the loss function with `reduction='none'` when using sample_weight.")
                    metrics[key] = (loss * sample_weight).mean().cpu().item()
                else:
                    metrics[key] = score_fn(y_pred, y_true).mean().cpu().item()

        return metrics

    @torch.no_grad()
    def evaluate(self, data_loader):
        self.model.eval()
    
        has_w = len(data_loader.dataset[0]) == 3

        metrics, y_true, y_pred = {}, [], []
        sample_weight = []

        for t in data_loader:
            x = t[0].to(device=self.device)
            y = t[1].to(device=self.device)
            y_true.append(y)
            y_pred.append(self.model.forward(x))

            if has_w:
                t[2] = t[2].to(device=self.device)
                sample_weight.append(t[2])

        y_true = torch.cat(y_true)
        y_pred = torch.cat(y_pred)

        if has_w:
            sample_weight = torch.cat(sample_weight)

        for key, score_fn in self.metrics.items():
            if has_w and 'loss' in key:
                loss = score_fn(y_pred, y_true)
                if loss.dim() == 0:
                    raise ValueError("You should set the loss function with `reduction='none'` when using sample_weight.")
                metrics[key] = (loss * sample_weight).mean().cpu().item()
            else:
                metrics[key] = score_fn(y_pred, y_true).mean().cpu().item()
        
        return metrics



class Logger(object):
    def __init__(self, trainer):
        self.trainer = trainer
        self.verbose = 0

    def on_train_begin(self, train_loader, val_loader):
        if self.verbose == 0:
            return None
        if val_loader != None:
            print(f'Train on {len(train_loader.dataset)} samples, validate on {len(val_loader.dataset)} samples:')
        else:
            print(f'Train on {len(train_loader.dataset)} samples:')

    def on_epoch_begin(self, **kwargs):
        self.time = time.time()

    def on_epoch_end(self, **kwargs):
        if not hasattr(self, 'history'):
            train_columns = [key for key in kwargs['train_metrics'].keys()]
            val_columns = ['val_'+key for key in kwargs['val_metrics'].keys()]
            self.ordered_metrics_keys = train_columns + val_columns
            self.history = pd.DataFrame(columns=self.ordered_metrics_keys)
        
        time_elapsed = time.time() - self.time

        content = []
        content.append(f"Epoch {kwargs['epoch']}/{kwargs['max_epochs']}")

        time_str = f'{round(time_elapsed, 1)}s' if time_elapsed < 10 else f'{int(time_elapsed + 0.5)}s'
        content.append(time_str)
        
        self.metrics = kwargs['train_metrics']
        self.metrics.update({('val_' + k): v for k, v in kwargs['val_metrics'].items()})
 
        for k in self.ordered_metrics_keys:
            content.append(f'{k}: ' + '{:.4f}'.format(self.metrics[k])) 

        self.metrics['lr'] = self.trainer.optimizer.param_groups[0]['lr']
        content.append('lr: {:.0e}'.format(self.metrics['lr']))

        self.history.loc[kwargs['epoch']] = self.metrics

        if self.verbose == 1:
            print(' - '.join(content))
        elif self.verbose == 2:
            print('|'.join(content).replace(' ', '').replace('Epoch', ''))


import torch
from torch.utils.data import DataLoader, TensorDataset
import torchsummary
from collections import OrderedDict
from torch.utils.data import random_split

from ._training import Trainer
from .layers import KerasLayer
from .metrics import Metric
from .metrics import create_metric_by_name
from .losses import create_loss_by_name
from .optimizers import create_optimizer_by_name
from .utils import to_tensor

__version__ = '0.6.2'

class Model(torch.nn.Module):
    """
    `Model` wraps a `nn.Module` with training and inference features.

    Once the model is wrapped, you can config the model with losses and metrics\n  with `model.compile()`, train the model with `model.fit()`, or use the model\n  to do prediction with `model.predict()`.
    """
    def __init__(self, model):
        super(Model, self).__init__()
        self.model = model
        self.compiled = False
        self.built = False

        self.has_keras_layer = False
        def check_keras_layer(m):
            if isinstance(m, KerasLayer):
                self.has_keras_layer = True
        self.model.apply(check_keras_layer)

    def forward(self, x):
        return self.model.forward(x)

    def count_params(self) -> int:
        """Count the total number of scalars composing the weights."""
        return sum([p.numel() for p in self.parameters()])

    ########## keras-style methods below ##########

    @torch.no_grad()
    def build(self, input_shape):
        """Build the model when it contains `KerasLayer`."""
        if self.has_keras_layer:
            input_shape = [2] + list(input_shape)
            probe_input = torch.zeros(size=input_shape)
            self.model.forward(probe_input)
        self.built = True
        return self

    def _check_keras_layer(self):
        if self.has_keras_layer and not self.built:
            raise AssertionError("You should call `model.build()` first because the model contains `KerasLayer`.")

    def summary(self, input_shape, depth=3):
        """Print a string summary of the network."""
        self._check_keras_layer()
        torchsummary.summary(self.model, input_shape, depth=depth, verbose=1)

    def compile(self, optimizer, loss, metrics=None, device=None):
        """
        Configure the model for training.

        Args:

        * `optimizer`: String (name of optimizer) or optimizer instance.

        * `loss`: String (name of objective function), objective function or loss instance.

        * `metrics`: List of metrics to be evaluated by the model during training. You can also use dict to specify the 
        abbreviation of each metric.

        * `device`: Device of the model and its trainer, if `None` 'cuda' will be used when `torch.cuda.is_available()` otherwise 'cpu'.
        """
        self._check_keras_layer()
        if device == None:
            device = 'cuda' if torch.cuda.is_available() else 'cpu'
            
        if isinstance(loss, str):
            loss = create_loss_by_name(loss)
        if isinstance(optimizer, str):
            optimizer = create_optimizer_by_name(optimizer, self.parameters())

        m = OrderedDict({'loss': loss})
        if isinstance(metrics, dict):
            m.update(metrics)
        elif isinstance(metrics, list):
            for tmp_m in metrics:
                if isinstance(tmp_m, str):
                    tmp_m = create_metric_by_name(tmp_m)
                if isinstance(tmp_m, Metric):
                    m[tmp_m.get_abbr()] = tmp_m
                    tmp_m.set_device(device)
                elif hasattr(tmp_m, '__call__'):
                    m[tmp_m.__name__] = tmp_m
                else:
                    raise TypeError('Unsupported type.')
        else:
            raise TypeError('Argument `metrics` should be either a dict or list.')

        self.to(device=device)
        self.trainer = Trainer(model=self, optimizer=optimizer, loss=loss, metrics=m, device=device)
        self.compiled = True


    def fit(self, x, y, epochs, batch_size=32,
                validation_split=None, val_split_seed=7,
                validation_data=None,
                callbacks=[],
                verbose=1,
                precise_train_metrics=False,
                shuffle=True,
                sample_weight=None
                ):
        """Train the model for a fixed number of epochs (iterations on a dataset)."""

        assert self.compiled
        x, y = to_tensor(x, y)

        assert not (validation_data != None and validation_split != None)
        has_val = validation_data != None or validation_split != None

        if type(sample_weight) != type(None):
            sample_weight = to_tensor(sample_weight).float()
            train_set = TensorDataset(x, y, sample_weight)
        else:
            train_set = TensorDataset(x, y)

        if validation_data != None:
            x_val, y_val = to_tensor(validation_data[0], validation_data[1])
            val_set = TensorDataset(x_val, y_val)

        if validation_split != None:
            val_length = int(len(train_set) * validation_split)
            train_length = len(train_set) - val_length
            train_set, val_set = random_split(train_set, [train_length, val_length], generator=torch.Generator().manual_seed(val_split_seed))

        train_loader = DataLoader(train_set, batch_size=batch_size, shuffle=shuffle)
        val_loader = DataLoader(val_set, batch_size=batch_size, shuffle=False) if has_val else None

        # Training
        self.trainer.register_callbacks(callbacks)
        history = self.trainer.run(train_loader, val_loader, max_epochs=epochs, verbose=verbose, precise_train_metrics=precise_train_metrics)

        return history

    @torch.no_grad()
    def evaluate(self, x, y, batch_size=32):
        """Return the loss value & metrics values for the model in test mode.\n\n    Computation is done in batches."""
        assert self.compiled
        x, y = to_tensor(x, y)
        val_loader = DataLoader(TensorDataset(x, y), batch_size=batch_size, shuffle=False)
        return self.trainer.evaluate(val_loader)

    @torch.no_grad()
    def predict(self, inputs, batch_size=32, device=None, output_numpy=True, activation=None):
        """Generate output predictions for the input samples.\n\n    Computation is done in batches."""
        self._check_keras_layer()

        if device == None:
            if self.compiled:
                device = self.trainer.device
            else:
                device = 'cuda' if torch.cuda.is_available() else 'cpu'

        inputs = to_tensor(inputs)
        self.eval().to(device=device)
        outputs = []

        data_loader = DataLoader(TensorDataset(inputs), batch_size=batch_size, shuffle=False)
        for x_batch in data_loader:
            outputs.append(self.forward(x_batch[0].to(device=device)))

        outputs = torch.cat(outputs, dim=0)

        if activation != None:
            outputs = activation(outputs)

        if output_numpy:
            return outputs.cpu().numpy()
        else:
            return outputs

    def save_weights(self, filepath):
        """Equal to `torch.save(model.state_dict(), filepath)`."""
        torch.save(self.state_dict(), filepath)

    def load_weights(self, filepath):
        """Equal to `model.load_state_dict(torch.load(filepath))`."""
        self.load_state_dict(torch.load(filepath))


from collections import OrderedDict
import torch.nn as nn

_losses_dict = OrderedDict({
    'mse': nn.MSELoss,
    'mae': nn.L1Loss,
    'categorical_crossentropy': nn.CrossEntropyLoss,
    'ce_loss': nn.CrossEntropyLoss,
    'binary_crossentropy': nn.BCEWithLogitsLoss,
    'bce_loss': nn.BCEWithLogitsLoss,
})

def create_loss_by_name(name):
    name = name.lower()
    if name not in _losses_dict:
        raise KeyError(f'Invalid loss name, we support {list(_losses_dict.keys())}.')
    return _losses_dict[name]()


import torch
from collections import OrderedDict

_optimizers_dict = OrderedDict({
    'adam': torch.optim.Adam,
    'rmsprop': torch.optim.RMSprop,
    'sgd': torch.optim.SGD
})

def create_optimizer_by_name(name, *args):
    name = name.lower()
    if name not in _optimizers_dict:
        raise KeyError(f'Invalid metric name, we support {list(_optimizers_dict.keys())}.')
    if name == 'sgd':
        lr = 1e-2
    else:
        lr = 1e-3
    return _optimizers_dict[name](*args, lr)


from collections import OrderedDict
import sklearn
import torch
import torch.nn.functional as F
from abc import abstractclassmethod

class Metric():
    def __init__(self) -> None:
        pass

    def set_device(self, device):
        self.device = device

    @abstractclassmethod
    def get_abbr(self) -> str:
        pass

class Accuracy(Metric):
    def __init__(self) -> None:
        super(Accuracy, self).__init__()

    def __call__(self, y_pred, y_true):
        if y_pred.shape[-1] == 1:
            return binary_accuracy(y_pred, y_true)
        else:
            return categorical_accuracy(y_pred, y_true)

    def get_abbr(self) -> str:
        return 'acc'

def categorical_accuracy(y_pred, y_true):
    return (y_pred.argmax(-1) == y_true).float().mean()

def binary_accuracy(y_pred, y_true, activation=torch.sigmoid):
    return (torch.round(activation(y_pred)) == y_true).float().mean()


class SklearnMetric(Metric):
    def __init__(self, activation=None) -> None:
        super(SklearnMetric, self).__init__()
        self.activation = activation

        try:
            import sklearn.metrics
        except ImportError:
            raise RuntimeError("This metric requires sklearn to be installed.")

        self.score_fn = self.get_score_fn()

    def __call__(self, y_pred, y_true):
        if self.activation != None:
            y_pred = self.activation(y_pred)
        y_true = y_true.cpu().numpy()
        y_pred = y_pred.cpu().numpy()
        return torch.tensor(self.score_fn(y_true, y_pred), dtype=torch.float32, device=self.device)

    @abstractclassmethod
    def get_score_fn(self):
        pass


class ROC_AUC(SklearnMetric):
    def __init__(self, activation=torch.sigmoid) -> None:
        super(ROC_AUC, self).__init__(activation=activation)

    def get_score_fn(self):
        return sklearn.metrics.roc_auc_score

    def get_abbr(self) -> str:
        return 'auc'

class F1_Score(SklearnMetric):
    def __init__(self, activation=torch.sigmoid) -> None:
        super(F1_Score, self).__init__(activation=lambda x: torch.round(activation(x)))

    def get_score_fn(self):
        return sklearn.metrics.f1_score

    def get_abbr(self) -> str:
        return 'f1'


# For Regression

class MeanSquaredError(Metric):
    def __init__(self) -> None:
        super(MeanSquaredError, self).__init__()

    def __call__(self, y_pred, y_true):
        return F.mse_loss(y_pred, y_true)

    def get_abbr(self) -> str:
        return 'mse'

class MeanAbsoluteError(Metric):
    def __init__(self) -> None:
        super(MeanAbsoluteError, self).__init__()

    def __call__(self, y_pred, y_true):
        return F.l1_loss(y_pred, y_true)

    def get_abbr(self) -> str:
        return 'mae'

class RootMeanSquaredError(Metric):
    def __init__(self) -> None:
        super(RootMeanSquaredError, self).__init__()

    def __call__(self, y_pred, y_true):
        return torch.sqrt(F.mse_loss(y_pred, y_true))

    def get_abbr(self) -> str:
        return 'rmse'


_metrics_dict = OrderedDict({
    'mse': MeanSquaredError,
    'mae': MeanAbsoluteError,
    'rmse': RootMeanSquaredError,
    'acc': Accuracy,
    'accuracy': Accuracy,
    'auc': ROC_AUC,
    'roc_auc': ROC_AUC,
    'f1': F1_Score,
    'f1_score': F1_Score,
})

def create_metric_by_name(name):
    name = name.lower()
    if name not in _metrics_dict:
        raise KeyError(f'Invalid metric name, we support {list(_metrics_dict.keys())}.')
    return _metrics_dict[name]()


from abc import abstractclassmethod
from ._training import Trainer, StopTrainingError, Events
import numpy as np

class Callback():
    def __init__(self) -> None:
        pass
    
    @abstractclassmethod
    def get_callbacks_dict(self):
        pass

def guess_auto_mode(monitor, mode) -> str:
    if mode != 'auto':
        assert mode == 'max' or mode == 'min'
        return mode
    for s in ['loss', 'mse', 'mae', 'mape', 'error', 'err']:
        if s in monitor:
            return 'min'
    return 'max'

class ModelCheckpoint(Callback):
    def __init__(self, filepath, monitor='val_loss', mode='auto', save_best_only=True, save_weights_only=True, verbose=0):
        super(ModelCheckpoint, self).__init__()

        if not (save_weights_only):
            raise ValueError('`ModelCheckpoint` only supports `save_weights_only=True`.')

        self.filepath = filepath
        self.monitor = monitor
        self.mode = guess_auto_mode(monitor, mode)
        self.save_best_only = save_best_only
        self.save_weights_only = save_weights_only
        self.verbose = verbose

        self.best_score = -np.inf if self.mode == 'max' else np.inf
        
    def is_better(self, curr_score):
        if self.mode == 'max':
            return curr_score > self.best_score
        else:   # mode == 'min'
            return curr_score < self.best_score

    def get_callbacks_dict(self):
        return {Events.ON_EPOCH_END: self.on_epoch_end}

    def on_epoch_end(self, trainer: Trainer):
        filename = self.filepath.format(epoch=trainer.epoch, **trainer.logger.metrics)

        if not self.save_best_only:
            trainer.model.save_weights(filename)
            return

        if self.monitor not in trainer.logger.metrics:
            raise KeyError(f'No such metric: {self.monitor}.')

        curr_score = trainer.logger.metrics[self.monitor]
        if self.is_better(curr_score):
            self.best_score = curr_score
            if self.verbose == 1:
                print("[INFO] Model saved at '{}'. The best score is {:.4f}.".format(self.filepath, self.best_score))
            trainer.model.save_weights(filename)


class EarlyStopping(Callback):
    def __init__(self, monitor='val_loss', mode='auto', min_delta=0, patience=5, baseline=None, verbose=1) -> None:
        super(EarlyStopping, self).__init__()

        self.monitor = monitor
        self.mode = guess_auto_mode(monitor, mode)
        self.min_delta = min_delta
        self.chances = self.patience = patience
        self.baseline = baseline
        self.verbose = verbose
        self.baseline_flag = False

        self.best_score = -np.inf if self.mode == 'max' else np.inf

    def is_better(self, curr_score):
        if self.mode == 'max':
            return curr_score - self.best_score > self.min_delta
        else:   # mode == 'min'
            if self.baseline != None and curr_score > self.baseline:
                return False
            return self.best_score - curr_score > self.min_delta

    def is_surpass_baseline(self, curr_score):
        if self.baseline == None or self.baseline_flag:
            return True
        if self.mode == 'max' and curr_score > self.baseline:
            self.baseline_flag = True
            return True
        if self.mode == 'min' and curr_score < self.baseline:
            self.baseline_flag = True
            return True
        return False

    def get_callbacks_dict(self):
        return {Events.ON_EPOCH_END: self.on_epoch_end}

    def on_epoch_end(self, trainer: Trainer):
        if self.monitor not in trainer.logger.metrics:
            raise KeyError(f'No such metric: {self.monitor}.')

        curr_score = trainer.logger.metrics[self.monitor]

        if not self.is_surpass_baseline(curr_score):
            return

        if self.is_better(curr_score):
            self.best_score = curr_score
            self.chances = self.patience
        else:
            self.chances -= 1
            if self.chances <= 0:
                if self.verbose == 1:
                    print('[INFO] Early Stopped. The best score is {:.4f}.'.format(self.best_score))
                raise StopTrainingError




class LRScheduler(Callback):
    def __init__(self, lr_scheduler) -> None:
        super(LRScheduler, self).__init__()
        self.lr_scheduler = lr_scheduler

    def get_callbacks_dict(self):
        return {Events.ON_EPOCH_END: self.on_epoch_end}

    def on_epoch_end(self, trainer):
        self.lr_scheduler.step()



class LambdaCallback(Callback):
    def __init__(self, on_epoch_begin=None, on_epoch_end=None, on_train_begin=None, on_train_end=None) -> None:
        super(LambdaCallback, self).__init__()

        self.callbacks_dict = {}

        if on_epoch_begin:
            self.callbacks_dict.update({Events.ON_EPOCH_BEGIN: on_epoch_begin})
        if on_epoch_end:
            self.callbacks_dict.update({Events.ON_EPOCH_END: on_epoch_end})
        if on_train_begin:
            self.callbacks_dict.update({Events.ON_TRAIN_BEGIN: on_train_begin})
        if on_train_end:
            self.callbacks_dict.update({Events.ON_TRAIN_END: on_train_end})

    def get_callbacks_dict(self):
        return self.callbacks_dict





import torch
import pandas as pd
import numpy as np

def to_tensor(*args):
    rt = []
    for arg in args:
        if isinstance(arg, pd.DataFrame):
            arg = torch.from_numpy(arg.values)
        elif isinstance(arg, np.ndarray):
            arg = torch.from_numpy(arg)
        elif isinstance(arg, list):
            arg = torch.tensor(arg)
        elif not isinstance(arg, torch.Tensor):
            raise TypeError('Only DataFrame, ndarray and torch.Tensor are supported.')

        if arg.dtype == torch.float64:
            arg = arg.float()
        elif arg.dtype == torch.int32:
            arg = arg.long()

        rt.append(arg)
                
    return rt[0] if len(rt) == 1 else tuple(rt)


import setuptools

with open("README.md", "rt", encoding='utf-8') as fh:
    long_description = fh.read()

setuptools.setup(
    name="keras4torch",
    version="0.6.2",
    author="Keras4Torch Team",
    author_email="blueloveTH@foxmail.com",
    description="An Easy-to-Use Wrapper for Training PyTorch Models",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/blueloveTH/keras4torch",
    packages=setuptools.find_packages(),
    install_requires=['torch-summary==1.4.3', 'pandas'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)